from chastack_bdd.tipos import *
from chastack_bdd.utiles import *
from chastack_bdd.usuario.errores import *
from chastack_bdd import Tabla, Registro
from chastack_bdd import ProtocoloBaseDeDatos, TipoCondicion, ErrorMalaSolicitud
from typing import Optional
import os
from secrets import token_bytes, token_urlsafe
from hashlib import sha256

class Usuario(Registro):
    """
    Clase base con funcionalidad común a usuarios.
    Maneja autenticación, encriptado de contraseñas,
    y gestión de sesiones y códigos de acceso únicos.
    """
    __slots__ = (
        "nombre_usuario",
        "correo",
        "contrasena",
        "id_sesion",
        "codigo_unico",
        "sal",
        "__fecha_ultimo_ingreso",
    )

    nombre_usuario : str
    correo : str
    contrasena : bytes
    __fecha_ultimo_ingreso : datetime
    id_sesion : str
    codigo_unico : Optional[str]
    sal : bytes

    @sobrecargar
    def __init__(
        self, 
        bdd : ProtocoloBaseDeDatos, 
        correo : str,
        contrasena : str,
        nombre_usuario : str = None
        ,*,debug : bool = False,
    ):
        correo = correo
        nombre_usuario = nombre_usuario if nombre_usuario is not None else correo
        contrasena, sal = self.encriptarContraseña(contrasena)
        id_sesion = self.__generarIdSesion()
        codigo_unico = None
        fecha_ultimo_ingreso = datetime.now()
        super().__init__(
            bdd=bdd,
            valores = dict(
                correo = correo,
                nombre_usuario = nombre_usuario,
                contrasena = contrasena,
                sal = sal,
                id_sesion = id_sesion,
                codigo_unico = codigo_unico,
                fecha_ultimo_ingreso = fecha_ultimo_ingreso,
            )
        )

    @classmethod
    def registrar(
        cls, 
        bdd : ProtocoloBaseDeDatos, 
        correo : str,
        contrasena : str,
        nombre_usuario : str = None,
        **nominales,
    ):
        devolverAtributoPrivado(cls,'__inicializar')(bdd) # HACER: (Herni) Generalizar a todos los @classmethods

        este_usuario = cls(bdd, correo, contrasena, nombre_usuario)
        este_usuario.__init__(bdd, valores=dict(**nominales))
        return este_usuario

    @classmethod
    @sobrecargar
    def ingresar(
        cls,
        bdd: ProtocoloBaseDeDatos,
        nombre_usuario: str,
        contrasena: str
    ) -> 'Usuario':
        """Autentica un usuario usando nombre de usuario/correo y contraseña.  
        **Parametros:**  
            :param ProtocoloBaseDeDatos bdd: Conexión a la base de datos.  
            :param str nombre_usuario: Nombre de usuario o correo electrónico.  
            :param str contrasena: Contraseña en texto plano.    
        
        **Levanta:**  
            :param ErrorAccesoIncorrecto: Si las credenciales son incorrectas  
        
        **Retorna:**  
            :param Usuario: Instancia del usuario autenticado    
        ---  
        """
        devolverAtributoPrivado(cls,'__inicializar')(bdd) # HACER: (Herni) Generalizar a todos los @classmethods

        datos : Optional[Resultado]
        columnas : tuple[str] = cls.__devolverColumnas()
        with bdd as bdd:
            bdd.SELECT(cls.__name__,columnas)
            if '@' in nombre_usuario:                
                bdd.WHERE(TipoCondicion.IGUAL,correo=nombre_usuario)
            else:
                bdd.WHERE(TipoCondicion.IGUAL,nombre_usuario=nombre_usuario)
            datos = bdd.ejecutar().devolverUnResultado()

        if datos is not None and cls.__verificarContraseña(datos.get('contrasena'), contrasena, datos.get('sal')):
            este_usuario = cls(bdd,valores = datos)
            este_usuario.id_sesion = cls.__generarIdSesion()
            este_usuario._actualizarFechaIngreso()
            este_usuario.guardar()
            return este_usuario
        else:
            raise ErrorAccesoIncorrecto("Usuario, correo o contraseña incorrectos.")
    
    @classmethod
    @sobrecargar
    def ingresar(
        cls,
        bdd : ProtocoloBaseDeDatos,
        id_sesion : str
    ) -> 'Usuario':
        devolverAtributoPrivado(cls,'__inicializar')(bdd) # HACER: (Herni) Generalizar a todos los @classmethods

        datos : Optional[Resultado]
        columnas : tuple[str] = cls.__devolverColumnas()
        with bdd:
            datos = bdd.SELECT(cls.__name__,columnas)\
                    .WHERE(TipoCondicion.IGUAL,id_sesion=id_sesion)\
                    .ejecutar().devolverUnResultado()

        if datos is not None:
            este_usuario = cls(bdd,valores = datos)
            este_usuario._actualizarFechaIngreso()
            este_usuario.guardar()
            return este_usuario
        else:
            raise ErrorMalaSolicitud("No existe usuario con el id_sesion provisto.")
        
    def cerrarSesion(self) -> None:
        """
        Cierra la sesión del usuario eliminando su ID de sesión.
        """
        self.id_sesion = None
        self.guardar()
    
    def cambiarContraseña(self, contrasena_nueva: str) -> Self:
        """
        Cambia la contraseña del usuario.
        
        Args:
            contrasena_nueva: Nueva contraseña en texto plano
        """
        contrasena_encriptada, sal = self.encriptarContraseña(contrasena_nueva)
        self.contrasena = contrasena_encriptada
        self.sal = sal
        self.guardar()
        return self
    
    @staticmethod
    def encriptarContraseña(contrasena: str, sal: bytes = None) -> tuple[bytes, bytes]:
        """Encripta una contraseña usando SHA-256 con sal y pimienta.  
        **Precondición:**  
            Debe existir la variable de entrono PIMIENTA, que contenga la cadena única 
            y persistente utilizada para el encriptado.  
        **Parametros:**  
            :param str contrasena: Contraseña en texto plano.  
            :param bytes sal: al para la encriptación. Si es `None`, se genera. Por defecto: None.  
        
        **Retorna:**  
            :param tuple[bytes, bytes]: Tupla con (contraseña_encriptada, sal)    
        
        ---  
        """
        if sal is None:
            sal = token_bytes(32) 
            
        pimienta = os.environ.get('PIMIENTA', '').encode('utf-8')
        contrasena_base = contrasena.encode('utf-8')
        
        contrasena_encriptada = sha256(contrasena_base + sal + pimienta).digest()
        return contrasena_encriptada, sal
    
    @staticmethod
    def __verificarContraseña(
        contrasena_encriptada: bytes, 
        contrasena: str, 
        sal: bytes
    ) -> bool:
        """Verifica si una contraseña coincide con su versión encriptada.    
        **Precondición:**  
            Debe existir la variable de entrono PIMIENTA, que contenga la cadena única 
            y persistente utilizada para el encriptado.  
        **Parametros:**  
            :param bytes contrasena_encriptada: Contraseña encriptada almacenada.  
            :param str contrasena: Contraseña en texto plano a verificar.  
            :param bytes sal: Sal utilizada en el encriptado.  
        
        **Retorna:**  
            :param bool: `True` si la contraseña es correcta, `False` en caso contrario.    
        
        ---  
        """
        return contrasena_encriptada == Usuario.encriptarContraseña(contrasena, sal)[0]

    @staticmethod 
    def __generarIdSesion() -> str:
        return token_urlsafe(64)
        
    @classmethod
    def __devolverColumnas(cls) -> tuple[str]:
        """Devuelve los `__slots__` de `Usuario` como columnas para operar con SQL.  
        
        **Retorna:**  
            :param tuple[str]: columnas SQL.    
        ---  
        """
        return (atributoPublico(atr) for atr in cls.__slots__ if atr not in ('__bdd','__tabla'))
    
    def _actualizarFechaIngreso(self, fecha_ultimo_ingreso : datetime = datetime.now()):
        self.__fecha_ultimo_ingreso = fecha_ultimo_ingreso


from chastack_bdd.tipos import *
from chastack_bdd.utiles import *
from chastack_bdd.bdd import ProtocoloBaseDeDatos
from chastack_bdd.registro import Registro, RegistroIntermedio

class Tabla(type):
    def __new__(mcs, nombre, bases, atributos):
    
        bases_ext = bases
        if not any(issubclass(base, Registro) for base in bases) and nombre != 'Registro':
            bases_ext = (Registro,) + bases
        
        cls = super().__new__(mcs, nombre, bases_ext, atributos)
        
        asignarAtributoPrivado(cls,'__tabla',nombre)
        setattr(cls, "tabla", property(lambda cls : cls.__tabla))

        
        if not hasattr(cls, '__annotations__'):
            cls.__annotations__ = {}
        
        return cls

    def __init__(cls, nombre, ancestros, diccionario):
        cls.__INICIALIZADA = False
        cls.__DEBUG = lambda msj : None

    def __call__(cls, bdd: ProtocoloBaseDeDatos, *posicionales, **nominales): 
        if nominales and nominales.get("debug", False):
            cls.__DEBUG = lambda msj: print(f"[DEBUG] {msj.rstrip()}")
        cls.__DEBUG(f"Se llamó a la clase {cls.__qualname__}. Instanciando objeto.")
        cls.__inicializar(bdd)

        instancia = super().__call__(bdd, *posicionales, **nominales)
        asignarAtributoPrivado(instancia,"__bdd",bdd)
        cls.__DEBUG(f"---------\n{instancia}\n---------\n")
        return instancia
    
    def __str__(cls):
        if not cls.__INICIALIZADA: 
            return f"<Tabla {cls.__name__}>"
        filas = {atributoPublico(ll) : v for ll,v in cls.__annotations__.items()}      
        ll_max, v_max = max(len(str(ll)) for ll, _ in filas.items() ), max(len(str(v)) for _, v in filas.items() )
        tabla_str = f"┌{'─' * (ll_max + 2)}┐\n" \
                    + f"│ {cls.__name__:<{ll_max}} │\n" \
                    + f"├{'─' * (ll_max + 2)}┼{'─' * (v_max + 2)}┬{'─' * (v_max + 2)}┐\n" \
                    + "\n".join(f"│ {str(ll):<{ll_max}} │ {str(v):<{v_max}} │ {str(cls.__tipoSQLDesdePython(v)):<{v_max}} │" for ll, v in filas.items()) \
                    + f"\n└{'─' * (ll_max + 2)}┴{'─' * (v_max + 2)}┴{'─' * (v_max + 2)}┘\n" 
        return tabla_str
    def __inicializar(cls,bdd):
            cls.__DEBUG(f"{cls.__qualname__} {'ya' if cls.__INICIALIZADA else 'no'} estaba inicializada.")
            if cls.__INICIALIZADA: return
            
            cls.__DEBUG(f"Inicializando modelo para: {cls.__tabla}.")
            slots :list[str] = []        
            anotaciones : dict[str,type] = {}
            with bdd as bdd:
                resultados = bdd.DESCRIBE(cls.__tabla).ejecutar().devolverResultados()
            
                for columna in resultados:
                    nombre_campo = columna.get('Field')
                    es_clave = columna.get('Key') == "PRI"
                    es_auto = "auto_increment" in columna.get("Extra", "").lower() or "default_generated" in columna.get("Extra", "").lower() or "auto_generated" in columna.get("Extra", "").lower()
                    
                    nombre_attr = f"__{nombre_campo}" if es_clave or es_auto else nombre_campo
                    
                    tipo = cls.__resolverTipo(columna.get('Type'), nombre_campo)
                    
                    if nombre_attr not in cls.__slots__:
                        slots.append(nombre_attr)
                    anotaciones.update({
                        nombre_attr : tipo
                    })
                    
                    if es_clave or es_auto:
                        setattr(cls, nombre_campo, property(lambda self, nombre_=nombre_campo: devolverAtributoPrivado(self,nombre_)))
                    cls.__bdd = bdd
            cls.__slots__ = cls.__slots__ + tuple(slots)
            cls.__annotations__.update(anotaciones)
            cls.__INICIALIZADA = True
            cls.__DEBUG(f"---------\n{cls}\n---------\n")
            

    @classmethod
    def __resolverTipo(cls, tipo_sql: str, nombre_columna: Optional[str]) -> type:
        """
        Deduce y devuelve un tipo de Python en base al tipo declarado en MySQL para la columna.
        Si encuentra un ENUM, crea un enum de Python y lo guarda como una constante de la clase.
        
        Parámetros:
            :arg tipo_sql str: El tipo definido en MySQL
            :arg nombre_columna Optional[str]: El nombre de la columna (útil para enums)
        
        Devuelve:
            :arg tipo `type`: el tipo python correspondiente (o `Any`)
        """
        tipo_declarado: Optional[Match[AnyStr]] = match(r'([a-z]+)(\(.*\))?', tipo_sql.lower())
        if not tipo_declarado:
            return Any

        tipo_base: str = tipo_declarado.group(1)
        parametros: str = tipo_declarado.group(2) if tipo_declarado.group(2) else ""
        tipo_completo: str = tipo_base + parametros

        tipos: dict[str, type] = {
            'tinyint': int,
            'smallint': int,
            'mediumint': int,
            'int': int,
            'bigint': int,
            'float': float,
            'double': float,
            'decimal': Decimal,
            'datetime': datetime,
            'timestamp': datetime,
            'date': date,
            'time': time,
            'char': str,
            'varchar': str,
            'text': str,
            'mediumtext': str,
            'longtext': str,
            'tinytext': str,
            'boolean': bool,
            'bool': bool,
            'tinyint(1)': bool,
            'blob': bytearray,
            'mediumblob': bytearray,
            'longblob': bytearray,
            'tinyblob': bytearray,
            'binary': bytes,
            'varbinary': bytes,
            'json': dict,
        }

        
        if tipo_base == 'enum':
            valores_enum: list[Any] = findall(r"'([^']*)'", tipo_sql)
            dicc_enum: dict[str, int] = {'_invalido': 0}
            for i, val in enumerate(valores_enum, 1):   
                dicc_enum[val] = i
            
            nombre_enum: str = f"Tipo{nombre_columna.capitalize()}" if nombre_columna else f"__ENUM_{token_urlsafe(4)}"
            clase_enum: type = type(
                nombre_enum,
                (EnumSQL, Enum),
                dicc_enum
            )
            
            setattr(cls,nombre_enum,clase_enum)
            return clase_enum

        
        if tipo_completo in tipos:
            return tipos[tipo_completo]
        return tipos.get(tipo_base, Any)

    @classmethod
    def __tipoSQLDesdePython(cls, tipo_python: type) -> str:
        """
        Devuelve el tipo SQL correspondiente a un tipo de Python.
        Si el tipo es un Enum generado por __resolverTipo, intenta reconstruir el ENUM SQL.
        
        Parámetros:
            :arg tipo_python type: el tipo de Python (p. ej., int, str, Enum, etc.)

        Devuelve:
            :arg str: el tipo SQL correspondiente
        """
        # Mapeo inverso de tipos básicos
        tipos_inversos: dict[type, str] = {
            int: 'int',
            float: 'double',
            Decimal: 'decimal(10,2)',
            datetime: 'timestamp',
            date: 'date',
            time: 'time',
            str: 'varchar(255)',
            bool: 'tinyint(1)',
            bytes: 'varbinary(255)',
            bytearray: 'blob',
            dict: 'json',
        }

        # Enums definidos dinámicamente por __resolverTipo
        if isinstance(tipo_python, type) and issubclass(tipo_python, EnumSQL):
            valores = [f"'{e.name}'" for e in tipo_python if e.name != '_invalido']
            return f"enum({','.join(valores)})"

        return tipos_inversos.get(tipo_python, 'text').upper()


    
class TablaIntermedia(type):
    def __new__(mcs, nombre, bases, atributos, esIntermedia: bool =True):
        
        bases_ext = bases
        if not any(issubclass(base, RegistroIntermedio) for base in bases) and nombre != 'RegistroIntermedio':
            bases_ext = (RegistroIntermedio,) + bases
        
        cls = super().__new__(mcs, nombre, bases_ext, atributos)
        
        asignarAtributoPrivado(cls,'__tabla',nombre)
        setattr(cls, "tabla", property(lambda cls : cls.__tabla))

        
        if not hasattr(cls, '__annotations__'):
            cls.__annotations__ = {}
        
        return cls

    def __init__(cls, nombre, ancestros, diccionario):
        cls.__INICIALIZADA = False
        cls.__DEBUG = lambda msj : None

    def __call__(cls, bdd: ProtocoloBaseDeDatos, *posicionales, **nominales): 
        if nominales and nominales.get("debug", False):
            cls.__DEBUG = lambda msj: print(f"[DEBUG] {msj.rstrip()}")
        cls.__DEBUG(f"Se llamó a la clase {cls.__qualname__}. Instanciando objeto.")
        cls.__inicializar(bdd)

        instancia = super().__call__(bdd, *posicionales, **nominales)
        asignarAtributoPrivado(instancia,"__bdd",bdd)
        cls.__DEBUG(f"---------\n{instancia}\n---------\n")
        return instancia
    
    def __str__(cls):
        if not cls.__INICIALIZADA: 
            return f"<TablaIntermedia {cls.__name__}>"
        filas = {atributoPublico(ll) : v for ll,v in cls.__annotations__.items()}      
        ll_max, v_max = max(len(str(ll)) for ll, _ in filas.items() ), max(len(str(v)) for _, v in filas.items() )
        tabla_str = f"┌{'─' * (ll_max + 2)}┐\n" \
                    + f"│ {cls.__name__:<{ll_max}} │\n" \
                    + f"├{'─' * (ll_max + 2)}┼{'─' * (v_max + 2)}┬{'─' * (v_max + 2)}┐\n" \
                    + "\n".join(f"│ {str(ll):<{ll_max}} │ {str(v):<{v_max}} │ {str(cls.__tipoSQLDesdePython(v)):<{v_max}} │" for ll, v in filas.items()) \
                    + f"\n└{'─' * (ll_max + 2)}┴{'─' * (v_max + 2)}┴{'─' * (v_max + 2)}┘\n" 
        return tabla_str
    def __inicializar(cls,bdd):
            cls.__DEBUG(f"{cls.__qualname__} {'ya' if cls.__INICIALIZADA else 'no'} estaba inicializada.")
            if cls.__INICIALIZADA: return
            
            cls.__DEBUG(f"Inicializando modelo para: {cls.__tabla}.")
            slots :list[str] = []        
            anotaciones : dict[str,type] = {}
            with bdd as bdd:
                resultados = bdd.DESCRIBE(cls.__tabla).ejecutar().devolverResultados()
            
                for columna in resultados:
                    nombre_campo = columna.get('Field')
                    es_clave = columna.get('Key') == "PRI"
                    es_auto = "auto_increment" in columna.get("Extra", "").lower() or "default_generated" in columna.get("Extra", "").lower() or "auto_generated" in columna.get("Extra", "").lower()
                    
                    nombre_attr = f"__{nombre_campo}" if es_clave or es_auto else nombre_campo
                    
                    tipo = cls.__resolverTipo(columna.get('Type'), nombre_campo)
                    
                    if nombre_attr not in cls.__slots__:
                        slots.append(nombre_attr)
                    anotaciones.update({
                        nombre_attr : tipo
                    })
                    
                    if es_clave or es_auto:
                        setattr(cls, nombre_campo, property(lambda self, nombre_=nombre_campo: devolverAtributoPrivado(self,nombre_)))
                    cls.__bdd = bdd
            cls.__slots__ = cls.__slots__ + tuple(slots)
            cls.__annotations__.update(anotaciones)
            cls.__INICIALIZADA = True
            cls.__DEBUG(f"---------\n{cls}\n---------\n")
            

    @classmethod
    def __resolverTipo(cls, tipo_sql: str, nombre_columna: Optional[str]) -> type:
        """
        Deduce y devuelve un tipo de Python en base al tipo declarado en MySQL para la columna.
        Si encuentra un ENUM, crea un enum de Python y lo guarda como una constante de la clase.
        
        Parámetros:
            :arg tipo_sql str: El tipo definido en MySQL
            :arg nombre_columna Optional[str]: El nombre de la columna (útil para enums)
        
        Devuelve:
            :arg tipo `type`: el tipo python correspondiente (o `Any`)
        """
        tipo_declarado: Optional[Match[AnyStr]] = match(r'([a-z]+)(\(.*\))?', tipo_sql.lower())
        if not tipo_declarado:
            return Any

        tipo_base: str = tipo_declarado.group(1)
        parametros: str = tipo_declarado.group(2) if tipo_declarado.group(2) else ""
        tipo_completo: str = tipo_base + parametros

        tipos: dict[str, type] = {
            'tinyint': int,
            'smallint': int,
            'mediumint': int,
            'int': int,
            'bigint': int,
            'float': float,
            'double': float,
            'decimal': Decimal,
            'datetime': datetime,
            'timestamp': datetime,
            'date': date,
            'time': time,
            'char': str,
            'varchar': str,
            'text': str,
            'mediumtext': str,
            'longtext': str,
            'tinytext': str,
            'boolean': bool,
            'bool': bool,
            'tinyint(1)': bool,
            'blob': bytearray,
            'mediumblob': bytearray,
            'longblob': bytearray,
            'tinyblob': bytearray,
            'binary': bytes,
            'varbinary': bytes,
            'json': dict,
        }

        
        if tipo_base == 'enum':
            valores_enum: list[Any] = findall(r"'([^']*)'", tipo_sql)
            dicc_enum: dict[str, int] = {'_invalido': 0}
            for i, val in enumerate(valores_enum, 1):   
                dicc_enum[val] = i
            
            nombre_enum: str = f"Tipo{nombre_columna.capitalize()}" if nombre_columna else f"__ENUM_{token_urlsafe(4)}"
            clase_enum: type = type(
                nombre_enum,
                (EnumSQL, Enum),
                dicc_enum
            )
            
            setattr(cls,nombre_enum,clase_enum)
            return clase_enum

        
        if tipo_completo in tipos:
            return tipos[tipo_completo]
        return tipos.get(tipo_base, Any)

    @classmethod
    def __tipoSQLDesdePython(cls, tipo_python: type) -> str:
        """
        Devuelve el tipo SQL correspondiente a un tipo de Python.
        Si el tipo es un Enum generado por __resolverTipo, intenta reconstruir el ENUM SQL.
        
        Parámetros:
            :arg tipo_python type: el tipo de Python (p. ej., int, str, Enum, etc.)

        Devuelve:
            :arg str: el tipo SQL correspondiente
        """
        # Mapeo inverso de tipos básicos
        tipos_inversos: dict[type, str] = {
            int: 'int',
            float: 'double',
            Decimal: 'decimal(10,2)',
            datetime: 'timestamp',
            date: 'date',
            time: 'time',
            str: 'varchar(255)',
            bool: 'tinyint(1)',
            bytes: 'varbinary(255)',
            bytearray: 'blob',
            dict: 'json',
        }

        # Enums definidos dinámicamente por __resolverTipo
        if isinstance(tipo_python, type) and issubclass(tipo_python, EnumSQL):
            valores = [f"'{e.name}'" for e in tipo_python if e.name != '_invalido']
            return f"enum({','.join(valores)})"

        return tipos_inversos.get(tipo_python, 'text').upper()

        